import sys
import logging
import re
import threading

from pkg_resources import iter_entry_points, load_entry_point

import tw
from tw.core.exceptions import WidgetException
from tw.core.util import LRUCache


__all__ = ["EngineManager", "Renderable", "EngineException", "display",
           "render"]


log = logging.getLogger(__name__)


class EngineException(WidgetException):
    pass

#XXX: This make_renderer/get_render_method/etc... is a mess which should be
#     cleaned sometime
def make_renderer(method, doc=None):
    def _renderer(self, renderable, **kw):
        template = kw.pop('template', renderable.template)
        if template is None:
            return
        origin = dynamic_best_engine(renderable, kw)
        destination = kw.get('displays_on', renderable.displays_on)
        if origin != 'cheetah':
            template = self.load_template(template, origin)
        renderer = self.get_render_method(origin, destination, method)
        output = renderer(info=kw, template=template)
        if method == 'display':
            output = self.adapt_output(output, destination)
        # Make sure string output is always unicode
        if isinstance(output, str):
            output = unicode(output, 'utf-8')
        return output
    _renderer.func_name = method
    _renderer.__doc__ = doc
    return _renderer

class EngineManager(dict):
    """
    Manages availble templating engines.
    """
    default_view = 'toscawidgets'

    def __init__(self, extra_vars_func=None, options=None, load_all=False):
        self.extra_vars_func = extra_vars_func
        self.options = options
        self._cache = LRUCache(50)
        self._lock = threading.Lock()
        if load_all:
            self.load_all()
            

    def __repr__(self):
        return "< %s >" % self.__class__.__name__


    def load_engine(self, name, options=None, extra_vars_func=None, 
                    distribution=None):
        factory = None
        if distribution:
            factory = load_entry_point(
                distribution, "python.templating.engines", name)
        else:
            for entrypoint in iter_entry_points("python.templating.engines"):
                if entrypoint.name == name:
                    factory = entrypoint.load()

        if factory is None:
            raise EngineException("No plugin available for engine '%s'" % name)

        options = options or self.options or {}
        options = options.copy()
        # emulate Kid and Genshi's dotted-path notation lookup
        options.setdefault('mako.directories', []).extend(sys.path)
        # make sure mako produces utf-8 output so we can decode it and use
        # unicode internally
        options['mako.output_encoding'] = 'utf-8'
        extra_vars_func = extra_vars_func or self.extra_vars_func

        self._lock.acquire()
        try:
            self[name] = factory(extra_vars_func, options)
        finally:
            self._lock.release()

    def __getitem__(self, name):
        """
        Return a Buffet plugin by name. If the plugin is not loaded it
        will try to load it with default arguments.
        """
        try:
            return dict.__getitem__(self, name)
        except KeyError:
            self.load_engine(name)
            return dict.__getitem__(self, name)


    def load_all(self, engine_options=None, stdvars=None, raise_error=False):
        for ep in iter_entry_points("python.templating.engines"):
            try:
                self.load_engine(ep.name, engine_options, stdvars)
            except:
                log.warn("Failed to load '%s' template engine: %r",
                         ep.name, sys.exc_info())
                if raise_error:
                    raise


    def load_template(self, template, engine_name):
        """Return's a compiled template and it's enginename"""
        output = None
        if isinstance(template, basestring) and _is_inline_template(template):
            # Assume inline template, try to fetch from cache
            key = (template, engine_name)
            try:
                output = self._cache[key]
                #log.debug("Cache hit for: %s", `key`)
            except KeyError:
                #log.debug("Cache miss for: %s", `key`)
                output = self._cache[key] = self[engine_name].load_template(
                    None, template_string = template
                    )
        elif isinstance(template, basestring):
            # Assume template path
            output = self[engine_name].load_template(template)
        else:
            # Assume compiled template
            output = template
        return output

    display = make_renderer('display', doc=\
        """
        Displays the renderable. Returns appropiate output for target template
        engine
        """)

    render = make_renderer('render', doc=\
        """
        Returns the serialized output in a string.
        Useful for debugging or to return to the browser as-is.
        """)

    def get_render_method(self, origin, destination, method):
        engine = self[origin]
        # Only pass-through Element/Stream output if rendering on the same
        # template engine as the one producing output
        if method == 'display' and \
          origin == destination and origin in ['kid', 'genshi']:
            return engine.transform
        # In any other case render as a string
        def _render_xhtml(**kw):
            kw.update(format='xhtml')
            return engine.render(**kw)
        return _render_xhtml


    def adapt_output(self, output, destination):
        # Handle rendering strings on Genshi so they're not escaped
        if isinstance(output, basestring) and destination == 'genshi':
            from genshi.input import HTML
            output = HTML(output)
        # Handle rendering strings on Kid so they're not escaped
        elif isinstance(output, basestring) and destination == 'kid':
            from kid import XML
            output = XML(output)
        return output
    
# Available as a public symbol for easier extension with PEAK-Rules for
# backwards compatibility
display = EngineManager.display.im_func
render = EngineManager.render.im_func

def choose_engine(obj, engine_name=None):
    tpl = obj.template
    if isinstance(tpl, basestring) and not _is_inline_template(tpl):
        colon = tpl.find(":")
        if colon > -1:
            engine_name = tpl[:colon]
            tpl = tpl[colon+1:]
    return engine_name, tpl

def dynamic_best_engine(renderable, params):
    try:
        ideal_engine = params['displays_on']
    except KeyError:
        ideal_engine = renderable.displays_on
    
    if ideal_engine not in renderable.available_engines:
        try:
            best_engine = params['engine_name']
        except KeyError:
            best_engine = renderable.engine_name
    else:
        best_engine = ideal_engine
    assert best_engine is not None
    return best_engine
    
_is_inline_template = re.compile(r'(<|\n|\$)').search

class Renderable(object):
    """Base class for all objects that the EngineManager can render"""

    engine_name = 'toscawidgets'
    available_engines = []
    template = None


    def displays_on(self):
        return tw.framework.default_view
    displays_on = property(displays_on, doc="""\
        Where the Renderable is being displayed on
        """)

    def __new__(cls, *args, **kw):
        obj = object.__new__(cls, *args, **kw)
        obj.template = kw.pop("template", obj.template)
        engine_name = kw.pop("engine_name", None)
        if obj.template is not None:
            colon_based_engine_name, obj.template = choose_engine(obj)
            # if there is a colon in the engine name, the available engines should be narrowed
            # to that engine.
            if colon_based_engine_name:
                obj.available_engines = [colon_based_engine_name,]
            engine_name = colon_based_engine_name
        if engine_name:
            obj.engine_name = engine_name
        return obj

    def render(self, **kw):
        kw.setdefault('_', tw.framework.translator)
        return tw.framework.engines.render(self, **kw)

    def display(self, **kw):
        kw.setdefault('_', tw.framework.translator)
        return tw.framework.engines.display(self, **kw)



"""
Default engine plugin for TGWidgets. 

Based on BuffetString which is Copyright (c) 2006 Christian Wyglendowski
"""

import string
import os

from pkg_resources import resource_filename

class ToscaWidgetsTemplatePlugin(object):
    # all template plugins need to define a default file extension
    extension = ".html"

    def __init__(self, extra_vars_func=None, config=None):
        """extra_vars_func == optional callable() that returns a dict
        config == optional dict() of configuration settings
        """
        self.get_extra_vars = extra_vars_func
        if config:
            self.config = config
        else:
            self.config = dict()

    def load_template(self, template_name, template_string=None):
        """
        template_name == dotted.path.to.template (without .ext)
        template_string == string containing the template
        
        The dotted notation is present because many template engines
        allow templates to be compiled down to Python modules.  TurboGears
        uses that feature to its adavantage, and for ease of integration
        the python.templating.engines plugin format requires the path to
        the template to be supplied as a dotted.path.to.template regardless
        of whether is is a module or not.

        In the case of string.Template templates, they are just simple text
        files, so we deal with the dotted notation and translate it into a
        standard file path to open the text file.
        """
        if template_string is not None:
            return string.Template(template_string)
        
        divider = template_name.rfind('.')
        if divider >= 0:
            package = template_name[:divider]
            basename = template_name[divider + 1:] + self.extension
            template_name = resource_filename(package, basename)

        template_file = open(template_name)
        template_obj = string.Template(template_file.read())
        template_file.close()
        
        return template_obj

    def render(self, info, format="html", fragment=False, template=None):
        """
        info == dict() of variables to stick into the template namespace
        format == output format if applicable
        fragment == special rules about rendering part of a page
        template == compiled template as returned by `load_template`
        """
        
        # check to see if we were passed a function get extra vars
        if callable(self.get_extra_vars):
            info.update(self.get_extra_vars())
        return template.safe_substitute(**info)



import os
import datetime
from pkg_resources import get_distribution, require
require("PasteScript")
from paste.script.templates import Template, var

class ToscaWidgetsTemplate(Template):
    _template_dir = os.path.join(get_distribution('ToscaWidgets').location,
                                 'tw', 'paste_templates', 'widget_template')
    summary = "ToscaWidgets widget"

    vars = [
        var('widget_name', 'Name of the widget package (tw.XXX)'),
        var('version', 'Version', default='0.1a0'),
        var('description', 'One-line description of the widget'),
        var('long_description', 'Multi-line description (in reST)'),
        var('author', 'Author name'),
        var('author_email', 'Author email'),
        var('url', 'URL of homepage'),
        var('license_name', 'License name'),
        ]

    def run(self, command, output_dir, vars):
        vars['year'] = str(datetime.datetime.now().year)
        vars['package'] = vars['widget_name'] or vars['package']
        super(ToscaWidgetsTemplate, self).run(command, output_dir, vars)



import logging
from decorator import decorator

from tw.api import retrieve_resources
from tw.mods.base import HostFramework
from tw.core.view import EngineManager

import pylons
from pylons.util import AttribSafeContextObj, ContextObj
from pylons.i18n import ugettext
from pylons.templating import render

from formencode import Invalid

__all__ = ["PylonsHostFramework", "validate", "render_response", "render",
           "valid"]

log = logging.getLogger(__name__)

class PylonsHostFramework(HostFramework):
    """HostFramework object for Pylons.

    Based on customization done in: http://wiki.pylonshq.com/display/pylonscookbook/An+Alternative+ToscaWidgets+Setup+with+Mako

    """
    def __init__(self, engines=None, default_view='mako', translator=ugettext,
                 template_paths=[], engine_options=None):
        if engines is None:
            opts = engine_options or {}
            opts.setdefault('mako.directories', template_paths)
            evf = opts.pop('extra_vars_func', None)
            engines = EngineManager(extra_vars_func=evf, options=opts)
        super(PylonsHostFramework, self).__init__(engines, default_view,
                                                  translator)
        



def validate(form=None, validators=None, error_handler=None, post_only=True, 
             state_factory=None):
    """This decorator will use valid() to automatically validate input.

    If validation is successful the decorated function will be called and the
    valid result dict will be saved as ``self.form_result``.
    
    Otherwise, the action will be re-run as if it was a GET without setting
    ``form_result`` and if the form is redisplayed it will display errors and
    previous input values.
    
    If the decorated method did not originally display the
    form, then ``error_handler`` should be the name of the method (in the same
    controller) that originally displayed it.
    
    If you'd like validate to also check GET (query) variables during its 
    validation, set the ``post_only`` keyword argument to False.
    """
    def wrapper(func, self, *args, **kwargs):
        """Decorator Wrapper function"""
        if not valid(self, form=form, validators=validators,
                     post_only=post_only,
                     state_factory=state_factory):
            if error_handler:
                environ = pylons.request.environ
                environ['REQUEST_METHOD'] = 'GET'
                environ['pylons.routes_dict']['action'] = error_handler
                return self._dispatch_call()
        return func(self, *args, **kwargs)
    return decorator(wrapper)


def valid(controller, form=None, validators=None, post_only=True,
          state_factory=None):
    """Validate input using a ToscaWidgetsForms form.
    
    Given a TW form or dict of validators, valid() will attempt to validate
    the form or validator dict as long as a POST request is made. No 
    validation is performed on GET requests unless post_only is False.
    
    If validation was succesfull, the valid result dict will be saved
    as ``controller.form_result`` and valid() will return True.

    Otherwise the invalid exception will be stored at
    ``controller.validation_exception`` and valid() will return False.

    If you'd like validate to also check GET (query) variables during its 
    validation, set the ``post_only`` keyword argument to False.
    """
    request = pylons.request._current_obj()

    if post_only:
        params = request.POST.copy()
    else:
        params = request.params.copy()

    errors = {}

    if state_factory:
        state = state_factory()
    else:
        from tw import framework
        # Pass registered translator for formencode
        state = type('State', (object,),
                     {'_':staticmethod(framework.translator)})

    if form:
        try:
            controller.form_result = form.validate(params, state=state)
        except Invalid, e:
            log.debug("Validation failed with:\n%s", e)
            controller.validation_exception = e
            errors = e.error_dict or e

    if validators:
        if isinstance(validators, dict):
            if not hasattr(controller, 'form_result'):
                controller.form_result = {}
            for field, validator in validators.iteritems():
                try:
                    controller.form_result[field] = \
                        validator.to_python(decoded[field] or None, state)
                except Invalid, error:
                    errors[field] = error
    if errors:
        controller.errors = errors
        return False
    return True


# Note: render and render_response are DEPRECATED

def _render_func_wrapper(func):
    def wrapper(*args, **kargs):
        import warnings
        warnings.warn(("%s is deprecated since collecting resources from "
                       "widgets at pylons.c.w is no longer needed to inject "
                       "them in the page.") % func.func_name,
                       DeprecationWarning, 2)
        from tw import framework
        if len(args) > 1 and args[0] in framework.engines:
            framework.default_view = args[0]

        global_widgets = getattr(pylons.g, 'w', None)
        request_widgets = getattr(pylons.c, 'w', None)
        other_resources = kargs.pop('resources', None)
        pylons.c.resources = retrieve_resources(
            [global_widgets, request_widgets, other_resources]
            )
        return func(*args, **kargs)

    try:
        wrapper.func_name = func.func_name
    except TypeError:
        # support 2.3
        pass
    tw_extra_doc = """\
This version is a ToscaWidgets wrapper which collects resources in
pylons.g.w and pylons.g.c and makies them available at pylons.c.resources
so the base template can render them.

It also sets the default_view if the engine name is overrided when calling me.
"""
    wrapper.__doc__ = func.__doc__ + "\n\n" + tw_extra_doc
    wrapper.__dict__ = func.__dict__
    return wrapper

render_response = _render_func_wrapper(pylons.templating.render_response)
render = _render_func_wrapper(pylons.templating.render)



import sys
import doctest

from new import instancemethod
from glob import glob
from unittest import TestCase
from itertools import imap, chain
from tw.core.util import install_framework

import pkg_resources

__all__ = [
    "RequireMixin", 
    "WidgetMixin", 
    "WidgetTestCase", 
    "WidgetRequireTestCase",
    "get_doctest_suite",
    ]

class RequireMixin(object):
    """
    Doesn't run the tests in the TestCases that inherit from this mixin
    class if the package requisites in 'require' are not met.
    """
    require = []
    def __init__(self, *args, **kw):
        try:
            pkg_resources.require(*self.require)
        except pkg_resources.DistributionNotFound:
            name = ':'.join([self.__class__.__module__, self.__class__.__name__])
            reqs = self.require
            self._message_displayed = False
            def dummy_run(self, result=None):
                if not self._message_displayed:
                    print >> sys.stderr, "Skipping all tests in %s due to missing " \
                                         "requirements: %r" % (name, reqs)
                    self._message_displayed = True
            self.run = instancemethod(dummy_run, self, self.__class__)
        TestCase.__init__(self, *args, **kw)
            
class WidgetMixin(object):
    widget_kw = {}
    def setUp(self):
        install_framework(force=True)
        if hasattr(self, 'TestWidget'):
            self.widget = self.TestWidget('test', **self.widget_kw)
    def tearDown(self):
        if hasattr(self, 'TestWidget'):
            del self.widget

    def assertInOutput(self, strings, *args, **kw):
        output = self.widget.render(*args, **kw)
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            self.failUnless(s in output, "%s\n\n%r not in output" %(output,s))

    def assertInOutputIC(self, strings, *args, **kw):
        output = self.widget.render(*args, **kw).lower()
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            s = s.lower()
            self.failUnless(s in output, "%s\n\n%r not in output" %(output,s))

    def assertNotInOutput(self, strings, *args, **kw):
        output = self.widget.render(*args, **kw)
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            self.failUnless(
                s not in output, "%s\n\n%r in output" %(output,s)
                )

    def assertNotInOutputIC(self, strings, *args, **kw):
        output = self.widget.render(*args, **kw).lower()
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            s = s.lower()
            self.failUnless(
                s not in output, "%s\n\n%r in output" %(output,s)
                )

    def assertInStaticCalls(self, strings):
        """Asserts given strings are included in the widget's static
        js calls."""
        from tw.core.resources import JSFunctionCalls
        calls = JSFunctionCalls(function_calls=self.widget._js_calls)
        output = calls.render()
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            self.failUnless(
                s in output, "%s\n\n%r not in static calls" %(output,s)
                )

    def assertInDynamicCalls(self, strings, *args, **kw):
        """Asserts given strings are included in the widget's dynamic
        js calls."""
        from tw.core.resources import dynamic_js_calls
        location = kw.pop('location', 'bodybottom')
        self.widget.render(*args, **kw)
        output = dynamic_js_calls.call_widgets[location].render()
        if isinstance(strings, basestring):
            strings = [strings]
        for s in strings:
            self.failUnless(
                s in output, "%s\n\n%r not in dynamic calls" %(output,s)
                )

class WidgetTestCase(WidgetMixin, TestCase):
    pass

class WidgetRequireTestCase(RequireMixin, WidgetTestCase):
    pass

def get_doctest_suite(doctest_files, doctest_modules):
    doctest_files = chain(*(imap(glob, doctest_files)))
    suite = doctest.DocFileSuite(
        *tuple(doctest_files),
        **{'optionflags':doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE,
           'module_relative':False,
           }
        )
    for mod in doctest_modules:
        try:
            suite.addTests(doctest.DocTestSuite(
                mod,
                **{'optionflags':doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE}
                ))
        except ValueError:
            pass # Mod has probably no doctests... ignore it
    return suite



from pkg_resources import require
require("CherryPy >=3.0")

import os, logging
import pprint

import cherrypy

import tw
from tw.core import resource_injector, resources
from tw.core.registry import Registry

log = logging.getLogger(__name__)

def _extract_config():
    c = cherrypy.config.copy()
    return c

class TWTool(cherrypy.Tool):
    """Sort-of-emulates TWWidgetsMiddleware + Paste's RegsitryManager. Takes
    care of preparing the hostframework for a request."""

    def __init__(self, host_framework, prefix='/toscawidgets',
                 serve_files=True):
        self.serve_files = serve_files
        self.prefix = prefix
        self.host_framework = host_framework
        
        return super(TWTool, self).__init__("on_start_resource", self.on_start_resource)

    def on_start_resource(self):
        log.debug("TWTool: on_start_resource")

        environ = cherrypy.request.wsgi_environ         
        registry = environ.setdefault('paste.registry', Registry())
        environ['toscawidgets.prefix'] = self.prefix
        registry.prepare()
        registry.register(tw.framework, self.host_framework)
        self.host_framework.start_request(environ)

    def before_request_body(self):
        """Intercepts requests for static files and serves them."""
        if not self.serve_files:
            return
        
        req, resp = cherrypy.request, cherrypy.response
        path = req.path_info
        if path.startswith(self.host_framework.webpath):
            path = path[len(self.host_framework.webpath):]
        
        if path.startswith(self.prefix):
            reg = resources.registry
            path = path[len(self.prefix)+len(reg.prefix):]
            stream, ct, enc = reg.get_stream_type_encoding(path)
            if stream:
                resp.body = stream
                if ct:
                    if enc:
                        ct += '; charset=' + enc
                    resp.headers['Content-Type'] = ct
                req.process_request_body = False
                req.handler = None

    def before_finalize(self):
        # Injects resources
        log.debug("TWTool: before_finalize")
        response = cherrypy.response
        ct = response.headers.get('content-type', 'text/html').lower()
        if 'html' in ct:
            cs = resource_injector.find_charset(ct)
            html = ''.join(response.body)
            resources = tw.framework.pop_resources()
            log.debug("Injecting Resources:")
            map(log.debug, pprint.pformat(resources).split('\n'))

            html = resource_injector.inject_resources(html=html,
                                                     resources=resources,
                                                     encoding=cs)
            response.body = [html]
            # Delete Content-Length header so finalize() recalcs it.
            response.headers.pop("Content-Length", None)

    def on_end_request(self):
        log.debug("TWTool: on_end_request")

        try:
            environ = cherrypy.request.wsgi_environ
            self.host_framework.end_request(environ)
        finally:
            registry = environ.get('paste.registry', None)
            if registry:
                registry.cleanup()
    
    def _setup(self):
        conf = self._merged_args()
        p = conf.pop("priority", None)
        if p is None:
            p = getattr(self.callable, "priority", self._priority)

        cherrypy.request.hooks.attach(self._point, self.callable, priority=p, **conf)
        cherrypy.request.hooks.attach('on_end_request', self.on_end_request)
        cherrypy.request.hooks.attach('before_request_body', self.before_request_body)
        cherrypy.request.hooks.attach('before_finalize', self.before_finalize)
        
def start_extension(host_framework, **filter_args):
    cherrypy.tools.toscawidgets = TWTool(host_framework=host_framework, **filter_args)



from tw.core.view import EngineManager
from tw.core.util import RequestLocalDescriptor, disable_runtime_checks
from tw.core.registry import StackedObjectProxy, Registry

__all__ = ["HostFramework"]

class RequestLocal(object):
    def __init__(self, environ):
        self.environ = environ
        self.resources = {}

class HostFramework(object):
    """
    This class is the interface between ToscaWidgets and the framework or
    web application that's using them.

    The an instance of this class should be passed as second argument to
    :class:`tw.core.middleware.ToscaWidgetsMiddleware` which will call its
    :meth:`start_request` method at the beginning of every
    request and :meth:`end_request` when the request is over so I have a chance
    to register our per-request context.

    A request-local proxy to a configured instance is placed at the beginning
    of the request at :attr:`tw.framework`

    **Constructor's arguments:**

    engines
       An instance of :class:`tw.core.viewEngineManager`.

    default_view
       The name of the template engine used by default in the container app's
       templates. It's used to determine what conversion is neccesary when
       displaying root widgets on a template. 

    translator
       Function used to translate strings.

    enable_runtime_checks
       Enables runtime checks for possible programming errors regarding
       modifying widget attributes once a widget has been initialized.
       Disabling this option can significantly reduce Widget initializatio
       time.

       .. note::
           This operation modifies the Widget class and will affect any
           application using ToscaWidgets in the same process.
    """
    request_local = StackedObjectProxy(name="ToscaWidgets per-request storage")
    request_local_class = RequestLocal

    default_view = RequestLocalDescriptor('default_view', 'toscawidgets')

    def __init__(self, engines=None, default_view='toscawidgets',
                 translator=lambda s: s, enable_runtime_checks=True, default_engine=None):
        if engines is None:
            engines = EngineManager()
        self.engines = engines
        self._default_view = default_view
        self._default_engine = default_engine
        if default_engine is None:
            self._default_engine = default_view
        self.translator = translator
        if not enable_runtime_checks:
            disable_runtime_checks()
    
    def start_request(self, environ):
        """
        Called by the middleware when a request has just begun so I have a
        chance to register the request context Widgets will use for various
        things.
        """
        registry = environ['paste.registry']
        registry.register(self.request_local, self.request_local_class(environ))
        self.request_local.default_view = self._default_view

    def end_request(self, environ):
        """
        Called by the middleware when a request has just finished so I can
        clean up.
        """
        pass

    def url(self, url):
        """
        Returns the absolute path for the given url.
        """
        prefix = self.request_local.environ['toscawidgets.prefix']
        script_name = self.request_local.environ['SCRIPT_NAME']
        return ''.join([script_name, prefix, url])

    def register_resources(self, resources):
        """
        Registers resources for injection in the current request.
        """
        from tw.api import merge_resources
        merge_resources(self.request_local.resources, resources)

    def pop_resources(self):
        """
        Returns returns the resources that have been registered for this
        request and removes them from request-local storage area.
        """
        resources =  self.request_local.resources
        self.request_local.resources = {}
        return resources
